#version 120

#define enlargementFactor 1.0
const float INFINITY = 1.0/0.0;

attribute vec4 in_position;
attribute vec3 in_normal;
attribute vec2 in_unitTexCoord, in_sizeTexCoord;
attribute vec4 in_color;

uniform mat4 u_model, u_view, u_projection;

uniform bool u_useAmbientProjection;
uniform float u_projectionPlaneOffset;
uniform vec4 u_normProjectionPlane;
uniform vec4 u_normAdditionalClipPlane;

uniform vec4 u_lightPositionWithRadius;

varying vec3 v_worldPosition;
varying float v_lightRadius; // half penumbra width
varying vec2 v_unitTexCoord, v_constCastPlaneSize;
varying vec4 v_color;

vec4 projectPointToPlaneAlong(vec3 point, vec4 nPlane, vec3 nDirection); // w-coord is distance


void main()
{
    vec3 basePoint = (u_model * in_position).xyz;
    vec3 lightPoint = u_lightPositionWithRadius.xyz;

    mat4 viewProjection = u_projection * u_view;
    float lightRadius = u_lightPositionWithRadius.w;
    //lightRadius = min(lightRadius, length(in_sizeTexCoord) / 2.0);


    vec4 modelSpaceTowardPlaneCenter = vec4(in_unitTexCoord * vec2(-2.0, 2.0) + vec2(1.0, -1.0), 0.0, 0.0);
    modelSpaceTowardPlaneCenter.y *= in_sizeTexCoord.x / in_sizeTexCoord.y; // Adjust for aspect ratio
    vec3 vecToPlaneCenter = (u_model * (in_position + modelSpaceTowardPlaneCenter)).xyz - basePoint; // Adjust to world space

    v_unitTexCoord = in_unitTexCoord;
    vec3 bestShadowPoint;

    if (u_useAmbientProjection) {
        // Expand by twice lightRadius so penumbra starts outside item
        vec3 expandedBasePoint = basePoint - length(vec2(2.0 * lightRadius)) * normalize(vecToPlaneCenter);
        vec3 lightDir = -u_normProjectionPlane.xyz;
        vec4 projectToShadowPlane = projectPointToPlaneAlong(expandedBasePoint, u_normProjectionPlane, lightDir);

        bestShadowPoint = projectToShadowPlane.xyz;
        v_lightRadius = lightRadius; // Why this fudge factor?

    } else {
        vec3 planeCenter = basePoint + normalize(vecToPlaneCenter) * (length(in_sizeTexCoord.xy) / 2.0);
        float radiusToDimension = (lightRadius * 2.0 * enlargementFactor) / max(in_sizeTexCoord.x, in_sizeTexCoord.y);
        float moveLightCloserBy = (radiusToDimension / (1.0 + radiusToDimension));
        vec3 adjustedLightPoint = lightPoint + moveLightCloserBy * (planeCenter - lightPoint);
        vec3 lightDir = normalize(basePoint - adjustedLightPoint);

        vec4 projectToShadowPlane = projectPointToPlaneAlong(basePoint, u_normProjectionPlane, lightDir);
        vec4 projectToClipPlane   = projectPointToPlaneAlong(basePoint, u_normAdditionalClipPlane, lightDir);

        bestShadowPoint = projectToShadowPlane.xyz;
        if (abs(projectToClipPlane.w) < abs(projectToShadowPlane.w)) {
            // Project clipped point down to shadow plane
            // WRONG: This drops it straight down, which is incorrect if plane has x-axis rotation
            bestShadowPoint = projectPointToPlaneAlong(projectToClipPlane.xyz, u_normProjectionPlane, u_normProjectionPlane.xyz).xyz;

            // This point now represents a different part of the book, so we need to adjust the texture coords
            // Project this new point back onto the book plane...
            vec4 sourcePlane = vec4(in_normal, -dot(in_normal, basePoint));
            vec4 representedSourcePoint = projectPointToPlaneAlong(bestShadowPoint, sourcePlane, normalize(bestShadowPoint - adjustedLightPoint));

            v_unitTexCoord += (basePoint.xy - representedSourcePoint.xy) / in_sizeTexCoord.xy; // (more assumptions here)
        }

        float planePointDistFromCastingPlane = dot(basePoint - bestShadowPoint, in_normal);
        v_lightRadius = max(0.0, lightRadius * planePointDistFromCastingPlane / distance(basePoint, lightPoint));
    }

    v_worldPosition = bestShadowPoint + u_projectionPlaneOffset * u_normProjectionPlane.xyz;

    v_constCastPlaneSize = in_sizeTexCoord + vec2(v_lightRadius * 2.0);

    v_color = in_color;
    gl_Position = viewProjection * vec4(v_worldPosition, 1.0);
}

vec4 projectPointToPlaneAlong(vec3 point, vec4 nPlane, vec3 nDirection)
{
    float projectedPointDistFromPoint = -(dot(point, nPlane.xyz) + nPlane.w) / dot(nDirection, nPlane.xyz);
    vec3 projectedPoint = point + projectedPointDistFromPoint * nDirection;
    return vec4(projectedPoint, projectedPointDistFromPoint);
}
